#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef USE_MATHLIB
#include <math.h>
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "const.h"
#include "types.h"
#include "nlllib.h"
#include "value.h"
#include "area.h"
#include "stack.h"
#include "function.h"
#include "libprintf.h"
#include "key.h"
#include "mouse.h"
#include "framebuf.h"
#include "graphic.h"

static int function_value(value_t value, value_t args);
static int function_true(value_t value, value_t args);
static int function_false(value_t value, value_t args);
static int function_cond(value_t value, value_t args);
static int function_at(value_t value, value_t args);
static int function_fpush(value_t value, value_t args);
static int function_fpop(value_t value, value_t args);
static int function_stacknum(value_t value, value_t args);
static int function_fprint(value_t value, value_t args);
static int function_sprintf(value_t value, value_t args);
static int function_fprintf(value_t value, value_t args);
static int function_fprintm(value_t value, value_t args);
static int function_input(value_t value, value_t args);
#ifdef NLL_STDIO_FUNCTION
static int function_fstdin(value_t value, value_t args);
static int function_fstdout(value_t value, value_t args);
#endif
static int function_itoc(value_t value, value_t args);
static int function_ctoi(value_t value, value_t args);
static int function_itoa(value_t value, value_t args);
static int function_atoi(value_t value, value_t args);
#ifdef NLL_FLOATING_POINT
static int function_itof(value_t value, value_t args);
static int function_ftoi(value_t value, value_t args);
#endif
static int function_length(value_t value, value_t args);
static int function_left(value_t value, value_t args);
static int function_right(value_t value, value_t args);
static int function_mid(value_t value, value_t args);
static int function_instr(value_t value, value_t args);
static int function_strcpy(value_t value, value_t args);
static int function_srand(value_t value, value_t args);
static int function_rand(value_t value, value_t args);
static int function_int(value_t value, value_t args);
static int function_abs(value_t value, value_t args);
static int function_sign(value_t value, value_t args);
static int function_time(value_t value, value_t args);
static int function_date(value_t value, value_t args);
static int function_memory(value_t value, value_t args);
static int function_peek(value_t value, value_t args);
static int function_poke(value_t value, value_t args);
static int function_read(value_t value, value_t args);
static int function_write(value_t value, value_t args);
static int function_memcpy(value_t value, value_t args);
#ifdef USE_MATHLIB
static int function_pi(value_t value, value_t args);
static int function_radian(value_t value, value_t args);
static int function_degree(value_t value, value_t args);
static int function_sin(value_t value, value_t args);
static int function_cos(value_t value, value_t args);
static int function_tan(value_t value, value_t args);
static int function_asin(value_t value, value_t args);
static int function_acos(value_t value, value_t args);
static int function_atan(value_t value, value_t args);
static int function_sinh(value_t value, value_t args);
static int function_cosh(value_t value, value_t args);
static int function_tanh(value_t value, value_t args);
static int function_exp(value_t value, value_t args);
static int function_log(value_t value, value_t args);
static int function_log10(value_t value, value_t args);
static int function_sqrt(value_t value, value_t args);
static int function_atan2(value_t value, value_t args);
static int function_pow(value_t value, value_t args);
static int function_fmod(value_t value, value_t args);
static int function_infinity(value_t value, value_t args);
static int function_nan(value_t value, value_t args);
static int function_isfinite(value_t value, value_t args);
static int function_isinf(value_t value, value_t args);
static int function_isnan(value_t value, value_t args);
static int function_isnormal(value_t value, value_t args);
#endif
static int function_fopen(value_t value, value_t args);
static int function_fclose(value_t value, value_t args);
static int function_fgets(value_t value, value_t args);
static int function_fputs(value_t value, value_t args);
static int function_fread(value_t value, value_t args);
static int function_fwrite(value_t value, value_t args);
static int function_keyname(value_t value, value_t args);
static int function_keycode(value_t value, value_t args);
static int function_screen(value_t value, value_t args);
static int function_scrstat(value_t value, value_t args);
static int function_flush(value_t value, value_t args);
static int function_getcolor(value_t value, value_t args);
static int function_setcolor(value_t value, value_t args);
static int function_getattr(value_t value, value_t args);
static int function_setattr(value_t value, value_t args);
static int function_getpos(value_t value, value_t args);
static int function_move(value_t value, value_t args);
static int function_getch(value_t value, value_t args);
static int function_clear(value_t value, value_t args);
static int function_scroll(value_t value, value_t args);
static int function_line(value_t value, value_t args);
static int function_box(value_t value, value_t args);
static int function_inkey(value_t value, value_t args);
static int function_gscreen(value_t value, value_t args);
static int function_gscrstat(value_t value, value_t args);
static int function_gcolor(value_t value, value_t args);
static int function_grgb(value_t value, value_t args);
static int function_gsync(value_t value, value_t args);
static int function_gflush(value_t value, value_t args);
static int function_ggetsize(value_t value, value_t args);
static int function_ggetcolor(value_t value, value_t args);
static int function_gsetcolor(value_t value, value_t args);
static int function_ggetpixel(value_t value, value_t args);
static int function_gclear(value_t value, value_t args);
static int function_gscroll(value_t value, value_t args);
static int function_gdot(value_t value, value_t args);
static int function_gline(value_t value, value_t args);
static int function_gbox(value_t value, value_t args);
static int function_gcircle(value_t value, value_t args);
static int function_gcopyimage(value_t value, value_t args);
static int function_ginkey(value_t value, value_t args);
static int function_gkeystat(value_t value, value_t args);
static int function_gmousepos(value_t value, value_t args);
static int function_ginmouse(value_t value, value_t args);
static int function_gmousestat(value_t value, value_t args);
static int function_gmainimage(value_t value, value_t args);
static int function_gallocimage(value_t value, value_t args);
static int function_gfreeimage(value_t value, value_t args);
static int function_gloadimage(value_t value, value_t args);
static int function_gsaveimage(value_t value, value_t args);

static int function_default_retnum(value_t args)
{
  int n = 0;
  for (; args; args = args->next)
    n++;
  n = (n > 0) ? n : 1;
  return n;
}

struct function functions[] = {
  { "VALUE"      , "FARGS"  , FUNCTION_TYPE_VALUE      , 1, function_default_retnum, function_value },
  { "TRUE"       , NULL     , FUNCTION_TYPE_TRUE       , 1, function_default_retnum, function_true },
  { "FALSE"      , NULL     , FUNCTION_TYPE_FALSE      , 1, function_default_retnum, function_false },
  { "COND"       , NULL     , FUNCTION_TYPE_COND       , 1, NULL, function_cond        },
  { "AT"         , NULL     , FUNCTION_TYPE_AT         , 1, NULL, function_at          },
  { "FPUSH"      , "LOCAL"  , FUNCTION_TYPE_FPUSH      , 1, function_default_retnum, function_fpush },
  { "FPOP"       , "RESTORE", FUNCTION_TYPE_FPOP       , 1, function_default_retnum, function_fpop },
  { "STACKNUM"   , NULL     , FUNCTION_TYPE_STACKNUM   , 1, NULL, function_stacknum    },
  { "FPRINT"     , NULL     , FUNCTION_TYPE_FPRINT     , 1, NULL, function_fprint      },
  { "SPRINTF"    , NULL     , FUNCTION_TYPE_SPRINTF    , 2, NULL, function_sprintf     },
  { "FPRINTF"    , NULL     , FUNCTION_TYPE_FPRINTF    , 1, NULL, function_fprintf     },
  { "FPRINTM"    , NULL     , FUNCTION_TYPE_FPRINTM    , 1, NULL, function_fprintm     },
  { "INPUT"      , NULL     , FUNCTION_TYPE_INPUT      , 1, NULL, function_input       },
#ifdef NLL_STDIO_FUNCTION
  { "FSTDIN"     , NULL     , FUNCTION_TYPE_FSTDIN     , 1, NULL, function_fstdin      },
  { "FSTDOUT"    , NULL     , FUNCTION_TYPE_FSTDOUT    , 1, NULL, function_fstdout     },
#endif
  { "ITOC"       , "CHAR"   , FUNCTION_TYPE_ITOC       , 1, function_default_retnum, function_itoc },
  { "CTOI"       , "ASCII"  , FUNCTION_TYPE_CTOI       , 1, function_default_retnum, function_ctoi },
  { "ITOA"       , "STR"    , FUNCTION_TYPE_ITOA       , 1, function_default_retnum, function_itoa },
  { "ATOI"       , "VAL"    , FUNCTION_TYPE_ATOI       , 1, function_default_retnum, function_atoi },
#ifdef NLL_FLOATING_POINT
  { "ITOF"       , NULL     , FUNCTION_TYPE_ITOF       , 1, function_default_retnum, function_itof },
  { "FTOI"       , NULL     , FUNCTION_TYPE_FTOI       , 1, function_default_retnum, function_ftoi },
#endif
  { "LENGTH"     , "LEN"    , FUNCTION_TYPE_LENGTH     , 1, function_default_retnum, function_length },
  { "LEFT"       , NULL     , FUNCTION_TYPE_LEFT       , 1, NULL, function_left        },
  { "RIGHT"      , NULL     , FUNCTION_TYPE_RIGHT      , 1, NULL, function_right       },
  { "MID"        , NULL     , FUNCTION_TYPE_MID        , 1, NULL, function_mid         },
  { "INSTR"      , NULL     , FUNCTION_TYPE_INSTR      , 1, NULL, function_instr       },
  { "STRCPY"     , NULL     , FUNCTION_TYPE_STRCPY     , 1, NULL, function_strcpy      },
  { "SRAND"      , NULL     , FUNCTION_TYPE_SRAND      , 1, NULL, function_srand       },
  { "RAND"       , "RND"    , FUNCTION_TYPE_RAND       , 1, function_default_retnum, function_rand },
  { "INT"        , NULL     , FUNCTION_TYPE_INT        , 1, function_default_retnum, function_int },
  { "ABS"        , NULL     , FUNCTION_TYPE_ABS        , 1, function_default_retnum, function_abs },
  { "SIGN"       , "SGN"    , FUNCTION_TYPE_SIGN       , 1, function_default_retnum, function_sign },
  { "TIME"       , NULL     , FUNCTION_TYPE_TIME       , 1, NULL, function_time        },
  { "DATE"       , NULL     , FUNCTION_TYPE_DATE       , 1, NULL, function_date        },
  { "MEMORY"     , NULL     , FUNCTION_TYPE_MEMORY     , 1, NULL, function_memory      },
  { "PEEK"       , NULL     , FUNCTION_TYPE_PEEK       , 1, NULL, function_peek        },
  { "POKE"       , NULL     , FUNCTION_TYPE_POKE       , 1, NULL, function_poke        },
  { "READ"       , NULL     , FUNCTION_TYPE_READ       , 1, NULL, function_read        },
  { "WRITE"      , NULL     , FUNCTION_TYPE_WRITE      , 1, NULL, function_write       },
  { "MEMCPY"     , NULL     , FUNCTION_TYPE_MEMCPY     , 1, NULL, function_memcpy      },
#ifdef USE_MATHLIB
  { "PI"         , NULL     , FUNCTION_TYPE_PI         , 1, function_default_retnum, function_pi },
  { "RADIAN"     , NULL     , FUNCTION_TYPE_RADIAN     , 1, function_default_retnum, function_radian },
  { "DEGREE"     , NULL     , FUNCTION_TYPE_DEGREE     , 1, function_default_retnum, function_degree },
  { "SIN"        , NULL     , FUNCTION_TYPE_SIN        , 1, function_default_retnum, function_sin },
  { "COS"        , NULL     , FUNCTION_TYPE_COS        , 1, function_default_retnum, function_cos },
  { "TAN"        , NULL     , FUNCTION_TYPE_TAN        , 1, function_default_retnum, function_tan },
  { "ASIN"       , NULL     , FUNCTION_TYPE_ASIN       , 1, function_default_retnum, function_asin },
  { "ACOS"       , NULL     , FUNCTION_TYPE_ACOS       , 1, function_default_retnum, function_acos },
  { "ATAN"       , NULL     , FUNCTION_TYPE_ATAN       , 1, function_default_retnum, function_atan },
  { "SINH"       , NULL     , FUNCTION_TYPE_SINH       , 1, function_default_retnum, function_sinh },
  { "COSH"       , NULL     , FUNCTION_TYPE_COSH       , 1, function_default_retnum, function_cosh },
  { "TANH"       , NULL     , FUNCTION_TYPE_TANH       , 1, function_default_retnum, function_tanh },
  { "EXP"        , NULL     , FUNCTION_TYPE_EXP        , 1, function_default_retnum, function_exp },
  { "LOG"        , NULL     , FUNCTION_TYPE_LOG        , 1, function_default_retnum, function_log },
  { "LOG10"      , NULL     , FUNCTION_TYPE_LOG10      , 1, function_default_retnum, function_log10 },
  { "SQRT"       , "SQR"    , FUNCTION_TYPE_SQRT       , 1, function_default_retnum, function_sqrt },
  { "ATAN2"      , NULL     , FUNCTION_TYPE_ATAN2      , 1, NULL, function_atan2       },
  { "POW"        , NULL     , FUNCTION_TYPE_POW        , 1, NULL, function_pow         },
  { "FMOD"       , NULL     , FUNCTION_TYPE_FMOD       , 1, NULL, function_fmod        },
  { "INFINITY"   , "INF"    , FUNCTION_TYPE_INFINITY   , 1, function_default_retnum, function_infinity },
  { "NAN"        , NULL     , FUNCTION_TYPE_NAN        , 1, function_default_retnum, function_nan },
  { "ISFINITE"   , NULL     , FUNCTION_TYPE_ISFINITE   , 1, function_default_retnum, function_isfinite },
  { "ISINF"      , NULL     , FUNCTION_TYPE_ISINF      , 1, function_default_retnum, function_isinf },
  { "ISNAN"      , NULL     , FUNCTION_TYPE_ISNAN      , 1, function_default_retnum, function_isnan },
  { "ISNORMAL"   , NULL     , FUNCTION_TYPE_ISNORMAL   , 1, function_default_retnum, function_isnormal },
#endif
  { "FOPEN"      , NULL     , FUNCTION_TYPE_FOPEN      , 1, NULL, function_fopen       },
  { "FCLOSE"     , NULL     , FUNCTION_TYPE_FCLOSE     , 1, NULL, function_fclose      },
  { "FGETS"      , NULL     , FUNCTION_TYPE_FGETS      , 1, NULL, function_fgets       },
  { "FPUTS"      , NULL     , FUNCTION_TYPE_FPUTS      , 1, NULL, function_fputs       },
  { "FREAD"      , NULL     , FUNCTION_TYPE_FREAD      , 1, NULL, function_fread       },
  { "FWRITE"     , NULL     , FUNCTION_TYPE_FWRITE     , 1, NULL, function_fwrite      },
  { "KEYNAME"    , NULL     , FUNCTION_TYPE_KEYNAME    , 1, function_default_retnum, function_keyname },
  { "KEYCODE"    , NULL     , FUNCTION_TYPE_KEYCODE    , 1, function_default_retnum, function_keycode },
  { "SCREEN"     , NULL     , FUNCTION_TYPE_SCREEN     , 1, NULL, function_screen      },
  { "SCRSTAT"    , NULL     , FUNCTION_TYPE_SCRSTAT    , 3, NULL, function_scrstat     },
  { "FLUSH"      , NULL     , FUNCTION_TYPE_FLUSH      , 1, NULL, function_flush       },
  { "GETCOLOR"   , NULL     , FUNCTION_TYPE_GETCOLOR   , 2, NULL, function_getcolor    },
  { "SETCOLOR"   , NULL     , FUNCTION_TYPE_SETCOLOR   , 1, NULL, function_setcolor    },
  { "GETATTR"    , NULL     , FUNCTION_TYPE_GETATTR    , 1, NULL, function_getattr     },
  { "SETATTR"    , NULL     , FUNCTION_TYPE_SETATTR    , 1, NULL, function_setattr     },
  { "GETPOS"     , NULL     , FUNCTION_TYPE_GETPOS     , 2, NULL, function_getpos      },
  { "MOVE"       , NULL     , FUNCTION_TYPE_MOVE       , 1, NULL, function_move        },
  { "GETCH"      , NULL     , FUNCTION_TYPE_GETCH      , 1, NULL, function_getch       },
  { "CLEAR"      , NULL     , FUNCTION_TYPE_CLEAR      , 1, NULL, function_clear       },
  { "SCROLL"     , NULL     , FUNCTION_TYPE_SCROLL     , 1, NULL, function_scroll      },
  { "LINE"       , NULL     , FUNCTION_TYPE_LINE       , 1, NULL, function_line        },
  { "BOX"        , NULL     , FUNCTION_TYPE_BOX        , 1, NULL, function_box         },
  { "INKEY"      , "KEY"    , FUNCTION_TYPE_INKEY      , 1, NULL, function_inkey       },
  { "GSCREEN"    , NULL     , FUNCTION_TYPE_GSCREEN    , 1, NULL, function_gscreen     },
  { "GSCRSTAT"   , NULL     , FUNCTION_TYPE_GSCRSTAT   , 4, NULL, function_gscrstat    },
  { "GCOLOR"     , NULL     , FUNCTION_TYPE_GCOLOR     , 1, NULL, function_gcolor      },
  { "GRGB"       , NULL     , FUNCTION_TYPE_GRGB       , 3, NULL, function_grgb        },
  { "GSYNC"      , NULL     , FUNCTION_TYPE_GSYNC      , 1, NULL, function_gsync       },
  { "GFLUSH"     , NULL     , FUNCTION_TYPE_GFLUSH     , 1, NULL, function_gflush      },
  { "GGETSIZE"   , NULL     , FUNCTION_TYPE_GGETSIZE   , 2, NULL, function_ggetsize    },
  { "GGETCOLOR"  , NULL     , FUNCTION_TYPE_GGETCOLOR  , 1, NULL, function_ggetcolor   },
  { "GSETCOLOR"  , NULL     , FUNCTION_TYPE_GSETCOLOR  , 1, NULL, function_gsetcolor   },
  { "GGETPIXEL"  , NULL     , FUNCTION_TYPE_GGETPIXEL  , 1, NULL, function_ggetpixel   },
  { "GCLEAR"     , NULL     , FUNCTION_TYPE_GCLEAR     , 1, NULL, function_gclear      },
  { "GSCROLL"    , NULL     , FUNCTION_TYPE_GSCROLL    , 1, NULL, function_gscroll     },
  { "GDOT"       , "GPOINT" , FUNCTION_TYPE_GDOT       , 1, NULL, function_gdot        },
  { "GLINE"      , NULL     , FUNCTION_TYPE_GLINE      , 1, NULL, function_gline       },
  { "GBOX"       , NULL     , FUNCTION_TYPE_GBOX       , 1, NULL, function_gbox        },
  { "GCIRCLE"    , NULL     , FUNCTION_TYPE_GCIRCLE    , 1, NULL, function_gcircle     },
  { "GCOPYIMAGE" , NULL     , FUNCTION_TYPE_GCOPYIMAGE , 1, NULL, function_gcopyimage  },
  { "GINKEY"     , "GKEY"   , FUNCTION_TYPE_GINKEY     , 1, NULL, function_ginkey      },
  { "GKEYSTAT"   , NULL     , FUNCTION_TYPE_GKEYSTAT   , 1, function_default_retnum, function_gkeystat },
  { "GMOUSEPOS"  , NULL     , FUNCTION_TYPE_GMOUSEPOS  , 2, NULL, function_gmousepos   },
  { "GINMOUSE"   , NULL     , FUNCTION_TYPE_GINMOUSE   , 3, NULL, function_ginmouse    },
  { "GMOUSESTAT" , NULL     , FUNCTION_TYPE_GMOUSESTAT , 1, function_default_retnum, function_gmousestat },
  { "GMAINIMAGE" , NULL     , FUNCTION_TYPE_GMAINIMAGE , 1, NULL, function_gmainimage  },
  { "GALLOCIMAGE", NULL     , FUNCTION_TYPE_GALLOCIMAGE, 1, NULL, function_gallocimage },
  { "GFREEIMAGE" , NULL     , FUNCTION_TYPE_GFREEIMAGE , 1, NULL, function_gfreeimage  },
  { "GLOADIMAGE" , NULL     , FUNCTION_TYPE_GLOADIMAGE , 3, NULL, function_gloadimage  },
  { "GSAVEIMAGE" , NULL     , FUNCTION_TYPE_GSAVEIMAGE , 1, NULL, function_gsaveimage  },
  { NULL         , NULL     , FUNCTION_TYPE_NONE       , 0, NULL, NULL                 }
};

const char *function_name(function_t function)
{
  return function->name;
}

/* #define NLL_FUNCTION_SHORTNAME */

int function_search(const char *name, function_t *functionp)
{
  function_t function;
#ifdef NLL_FUNCTION_SHORTNAME
  int len;
#endif

  for (function = functions; function->name; function++) {
    if (!strcmp(function->name, name)) {
      *functionp = function;
      return 0;
    }
  }

  for (function = functions; function->name; function++) {
    if (function->other_name && !strcmp(function->other_name, name)) {
      *functionp = function;
      return 0;
    }
  }

#ifdef NLL_FUNCTION_SHORTNAME
  for (len = 0; name[len]; len++) {
    if ((name[len] == '.') && !name[len + 1]) {
      for (function = functions; function->name; function++) {
	if (!strncmp(function->name, name, len)) {
	  *functionp = function;
	  return 0;
	}
      }
      break;
    }
  }
#endif

  return NLL_ERRCODE_FORMULA_UNKNOWN_FUNCTION;
}

int function_list(FILE *fp)
{
  function_t function;

  for (function = functions; function->name; function++) {
    fprintf(fp, "%s", function->name);
    if (function->other_name)
      fprintf(fp, "\t(%s)", function->other_name);
    fprintf(fp, "\n");
  }

  return 0;
}

int function_proc(function_t function, value_t value, value_t args)
{
  return function->function(value, args);
}

static int function_value(value_t value, value_t args)
{
  int r;

  for (; args && value; args = args->next, value = value->next) {
    if ((r = value_set_value(value, args)) < 0)
      return r;
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_true(value_t value, value_t args)
{
  int r;

  if (!args) {
    if (value) {
      if ((r = value_set_integer(value, 1)) < 0)
	return r;
      value = value->next;
    }
  } else {
    for (; args; args = args->next) {
      if (value) {
	if ((r = value_set_integer(value, 1)) < 0)
	  return r;
	value = value->next;
      }
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_false(value_t value, value_t args)
{
  int r;

  if (!args) {
    if (value) {
      if ((r = value_set_integer(value, 0)) < 0)
	return r;
      value = value->next;
    }
  } else {
    for (; args; args = args->next) {
      if (value) {
	if ((r = value_set_integer(value, 0)) < 0)
	  return r;
	value = value->next;
      }
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_cond(value_t value, value_t args)
{
  int r, n;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_is_nzero(args)) < 0)
    return r;
  n = r ? 0 : 1;
  args = args->next;

  for (; args && (n > 0); args = args->next, n--)
    ;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if (!args) {
    if ((r = value_clear(value)) < 0)
      return r;
  } else {
    if ((r = value_set_value(value, args)) < 0)
      return r;
  }

  return 0;
}

static int function_at(value_t value, value_t args)
{
  int r;
  integer_t integer = 0;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  args = args->next;

  for (; args && (integer > 0); args = args->next, integer--)
    ;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if (!args) {
    if ((r = value_clear(value)) < 0)
      return r;
  } else {
    if ((r = value_set_value(value, args)) < 0)
      return r;
  }

  return 0;
}

static int function_fpush(value_t value, value_t args)
{
  int r;

  for (; args; args = args->next) {
    if ((r = stack_push(args)) < 0)
      return r;
    if (value) {
      if ((r = value_set_value(value, args)) < 0)
	return r;
      value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_fpop(value_t value, value_t args)
{
  int r;

  if (!args) {
    if (value) {
      if ((r = stack_pop(value)) < 0)
	return r;
      value = value->next;
    }
  } else {
    for (; args; args = args->next) {
      if ((r = stack_pop(args)) < 0)
	return r;
      if (value) {
	if ((r = value_set_value(value, args)) < 0)
	  return r;
	value = value->next;
      }
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_stacknum(value_t value, value_t args)
{
  int r, n;

  n = stack_getnum();

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_fprint(value_t value, value_t args)
{
  int r, n = 0;

  value_output_values(args, NULL);

  for (; args; args = args->next, n++)
    ;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_sprintf(value_t value, value_t args)
{
  int r, n;
  char *format;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &format, NULL)) < 0)
    return r;
  args = args->next;

  n = nll_sprintf(buffer, sizeof(buffer), format, args);
  if (n < 0)
    strcpy(buffer, "");

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, -1)) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_fprintf(value_t value, value_t args)
{
  int r, n;
  char *format;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &format, NULL)) < 0)
    return r;
  args = args->next;

  n = nll_sprintf(buffer, sizeof(buffer), format, args);
  if (n > 0)
    framebuf_puts(buffer);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_fprintm(value_t value, value_t args)
{
  int r;
  integer_t x = -1, y = -1, fgcolor = -1, bgcolor = -1;
  value_t arg;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  arg = args;
  args = args->next;

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &fgcolor)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &bgcolor)) < 0)
      ;
  }

  framebuf_move_cursor(x, y);
  if (fgcolor >= 0) framebuf_setfgcolor(fgcolor);
  if (bgcolor >= 0) framebuf_setbgcolor(bgcolor);
  value_output_value(arg);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, 0)) < 0)
    return r;

  return 0;
}

static int function_input(value_t value, value_t args)
{
  int r, len;
  char *string;
  char buffer[STRING_MAXLEN + 1];

  framebuf_enable(0);

  for (; args; args = args->next) {
    if (value_get_string(args, &string, NULL) < 0)
      continue;
    fprintf(nll_stdout, "%s", string);
  }
  fflush(nll_stdout);

  if (!fgets(buffer, sizeof(buffer), nll_stdin))
    buffer[0] = '\0';

  len = strlen(buffer);
  while ((len > 0) && (buffer[len - 1] == '\n'))
    buffer[--len] = '\0';

  framebuf_enable(1);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, -1)) < 0)
    return r;

  return 0;
}

#ifdef NLL_STDIO_FUNCTION
static int function_fstdin(value_t value, value_t args)
{
  int r;
  char *filename = NULL;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &filename, NULL)) < 0)
    return r;

  if (!filename) {
    r = (nll_stdin_restore() < 0) ? -1 : 0;
  } else {
    fp = fopen(filename, "r");
    if (fp)
      r = (nll_stdin_save(fp) < 0) ? -1 : 0;
    else
      r = -1;
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
}

static int function_fstdout(value_t value, value_t args)
{
  int r;
  char *filename = NULL;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &filename, NULL)) < 0)
    return r;

  if (!filename) {
    r = (nll_stdout_restore() < 0) ? -1 : 0;
  } else {
    fp = fopen(filename, "w");
    if (fp)
      r = (nll_stdout_save(fp) < 0) ? -1 : 0;
    else
      r = -1;
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
}
#endif

static int function_itoc(value_t value, value_t args)
{
  int r;
  integer_t integer;
  char str[2];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
    str[0] = integer;
    str[1] = '\0';
    if (value) {
      if ((r = value_set_string(value, str, -1)) < 0)
	return r;
      value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_ctoi(value_t value, value_t args)
{
  int r;
  integer_t integer;
  char *string;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_string(args, &string, NULL)) < 0)
      return r;
    integer = *(unsigned char *)string;
    if (value) {
      if ((r = value_set_integer(value, integer)) < 0)
	return r;
      value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_itoa(value_t value, value_t args)
{
  int r;
  integer_t integer;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
    sprintf(buffer, "%ld", integer);
    if (value) {
      if ((r = value_set_string(value, buffer, -1)) < 0)
	return r;
      value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_atoi(value_t value, value_t args)
{
  int r;
  integer_t integer;
  char *string;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_string(args, &string, NULL)) < 0)
      return r;
    integer = (int)strtol(string, NULL, 0);
    if (value) {
      if ((r = value_set_integer(value, integer)) < 0)
	return r;
      value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

#ifdef NLL_FLOATING_POINT
static int function_itof(value_t value, value_t args)
{
  int r;
  integer_t integer;
  double f;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
    f = (double)integer;
    if (value) {
      if ((r = value_set_float(value, f)) < 0)
	return r;
      value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_ftoi(value_t value, value_t args)
{
  int r;
  integer_t integer;
  double f;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_float(args, &f)) < 0)
      return r;
    integer = (int)f;
    if (value) {
      if ((r = value_set_integer(value, integer)) < 0)
	return r;
      value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}
#endif

static int function_length(value_t value, value_t args)
{
  int r, length;
  char *string;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_string(args, &string, &length)) < 0)
      return r;
    if (value) {
      if ((r = value_set_integer(value, length)) < 0)
	return r;
     value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_left(value_t value, value_t args)
{
  int r, i, length;
  integer_t integer = 1;
  char *s, *p;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s, &length)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &integer)) < 0)
      ;
  }

  p = buffer;
  for (i = 0; (i < length) && (i < integer); i++) {
    *(p++) = *(s++);
  }
  *p = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, i)) < 0)
    return r;

  return 0;
}

static int function_right(value_t value, value_t args)
{
  int r, i, length;
  integer_t integer = 1;
  char *s, *p;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s, &length)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &integer)) < 0)
      ;
  }

  s += length - integer;
  p = buffer;
  for (i = 0; i < integer; i++) {
    *(p++) = *(s++);
  }
  *p = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, i)) < 0)
    return r;

  return 0;
}

static int function_mid(value_t value, value_t args)
{
  int r, i, length;
  integer_t integer0 = 1, integer1 = -1;
  char *s, *p;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s, &length)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &integer0)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &integer1)) < 0)
      ;
  }

  s += integer0;
  p = buffer;
  for (i = 0; (integer0 + i < length) && ((integer1 < 0) || (i < integer1)); i++) {
    *(p++) = *(s++);
  }
  *p = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, i)) < 0)
    return r;

  return 0;
}

static int function_instr(value_t value, value_t args)
{
  int r, i, n, length0, length1;
  char *s0, *s1;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s0, &length0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s1, &length1)) < 0)
    return r;

  n = -1;
  for (i = 0; i <= length0 - length1; i++) {
    if (!memcmp(s0 + i, s1, length1)) {
      n = i;
      break;
    }
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_strcpy(value_t value, value_t args)
{
  int r, length0, length1;
  integer_t offset = -1;
  char *s0, *s1;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s0, &length0)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s1, &length1)) < 0)
    return r;

  if (offset < 0)
    offset = length0;

  if (offset + length1 > STRING_MAXLEN)
    return NLL_ERRCODE_STRING_TOO_LONG;

  memcpy(buffer, s0, length0);
  memcpy(buffer + offset, s1, length1);

  if (offset + length1 > length0)
    length0 = offset + length1;

  buffer[length0] = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, length0)) < 0)
    return r;

  return 0;
}

static unsigned int seed = 1;

static unsigned int _rand(void)
{
  seed ^= seed << 13;
  seed ^= seed >> 17;
  seed ^= seed <<  5;
  return seed;
}

static int function_srand(value_t value, value_t args)
{
  int r;
  integer_t s;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &s)) < 0)
    return r;

  seed = !s ? time(NULL) : s;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, seed)) < 0)
    return r;

  return 0;
}

static int function_rand(value_t value, value_t args)
{
  int r, n;
  integer_t range = 0;

  if (!args) {
    _rand();
    if (value) {
      if ((r = value_set_integer(value, seed)) < 0)
	return r;
      value = value->next;
    }
  } else {
    for (; args; args = args->next) {
      if ((r = value_get_integer(args, &range)) < 0)
	return r;
      n = range;
      _rand();
      if (value) {
	if ((r = value_set_integer(value, (n > 0) ? (seed % n) : seed)) < 0)
	  return r;
	value = value->next;
      }
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_int(value_t value, value_t args)
{
  int r;
  integer_t integer;
#ifdef NLL_FLOATING_POINT
  double f;
#endif

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_integer(args, &integer)) < 0) {
#ifdef NLL_FLOATING_POINT
      if ((r = value_get_float(args, &f)) < 0)
	return r;
      integer = f;
      while (integer > f)
	integer--;
#else
      return r;
#endif
    }
    if (value) {
      if ((r = value_set_integer(value, integer)) < 0)
	return r;
      value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_abs(value_t value, value_t args)
{
  int r, is_float;
  integer_t integer;
#ifdef NLL_FLOATING_POINT
  double f;
#endif

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    is_float = 0;
    if ((r = value_get_integer(args, &integer)) < 0) {
#ifdef NLL_FLOATING_POINT
      if ((r = value_get_float(args, &f)) < 0)
	return r;
      f = (f < 0.0) ? -f : f;
      is_float = 1;
#else
      return r;
#endif
    } else {
      integer = (integer < 0) ? -integer : integer;
    }
    if (value) {
      if (is_float) {
#ifdef NLL_FLOATING_POINT
	if ((r = value_set_float(value, f)) < 0)
	  return r;
#else
	return NLL_ERRCODE_VALUE_INVALID_TYPE;
#endif
      } else {
	if ((r = value_set_integer(value, integer)) < 0)
	  return r;
      }
      value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_sign(value_t value, value_t args)
{
  int r, is_float;
  integer_t integer;
#ifdef NLL_FLOATING_POINT
  double f;
#endif

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    is_float = 0;
    if ((r = value_get_integer(args, &integer)) < 0) {
#ifdef NLL_FLOATING_POINT
      if ((r = value_get_float(args, &f)) < 0)
	return r;
      if      (f < 0.0) f = -1.0;
      else if (f > 0.0) f =  1.0;
      else              f =  0.0;
      is_float = 1;
#else
      return r;
#endif
    } else {
      if      (integer < 0) integer = -1;
      else if (integer > 0) integer =  1;
      else                  integer =  0;
    }
    if (value) {
      if (is_float) {
#ifdef NLL_FLOATING_POINT
	if ((r = value_set_float(value, f)) < 0)
	  return r;
#else
	return NLL_ERRCODE_VALUE_INVALID_TYPE;
#endif
      } else {
	if ((r = value_set_integer(value, integer)) < 0)
	  return r;
      }
      value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_time(value_t value, value_t args)
{
  int r, n;
  time_t t;

  t = time(NULL);
  n = t;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_date(value_t value, value_t args)
{
  int r;
#ifndef USE_NLLIBC
  time_t t;
#endif
  char *s;

#ifndef USE_NLLIBC
  t = time(NULL);
  s = ctime(&t);
#else
  s = "Cannot get date.";
#endif

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, s, -1)) < 0)
    return r;

  return 0;
}

static int function_memory(value_t value, value_t args)
{
  int r;
  integer_t integer;
  area_t area;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  if ((r = area_alloc(&area, integer)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_area(value, area)) < 0)
    return r;

  return 0;
}

static int function_peek(value_t value, value_t args)
{
  int r;
  integer_t offset = 0, size = -1, integer;
  area_t area;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
  }

  if ((r = area_read_integer(area, offset, &integer, size)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, integer)) < 0)
    return r;

  return 0;
}

static int function_poke(value_t value, value_t args)
{
  int r;
  integer_t offset = 0, size = -1, integer;
  area_t area;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
    args = args->next;
  }

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  if ((r = area_write_integer(area, offset, integer, size)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, integer)) < 0)
    return r;

  return 0;
}

static int function_read(value_t value, value_t args)
{
  int r;
  integer_t offset = 0, size = -1;
  area_t area;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
  }

  if (size < 0)
    size = area_get_size(area) - offset;

  if (size < 0)
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  if (size > STRING_MAXLEN)
    return NLL_ERRCODE_STRING_TOO_LONG;

  if ((r = area_read(area, offset, buffer, size)) < 0)
    return r;
  buffer[size] = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, size)) < 0)
    return r;

  return 0;
}

static int function_write(value_t value, value_t args)
{
  int r, length;
  integer_t offset = 0, size = -1;
  area_t area;
  char *string;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
    args = args->next;
  }

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &string, &length)) < 0)
    return r;

  if (size < 0)
    size = length;

  if (size > length)
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  if ((r = area_write(area, offset, string, size)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, size)) < 0)
    return r;

  return 0;
}

static int function_memcpy(value_t value, value_t args)
{
  int r;
  integer_t doffset = 0, soffset = 0, size = -1;
  area_t darea, sarea;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &darea)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &doffset)) < 0)
      ;
    args = args->next;
  }

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &sarea)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &soffset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
  }

  if (size < 0)
    size = area_get_size(sarea) - soffset;

  if (size < 0)
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  if ((r = area_copy_area(darea, doffset, sarea, soffset, size)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, size)) < 0)
    return r;

  return 0;
}

#ifdef USE_MATHLIB
static int function_mathfunc(function_type_t type, value_t value, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  double f;

  if (!args) {
    switch (type) {
    case FUNCTION_TYPE_PI: f = M_PI; break;
    default:
      return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
    }
    if (value) {
      if ((r = value_set_float(value, f)) < 0)
	return r;
      value = value->next;
    }
  } else {
    for (; args; args = args->next) {
      if ((r = value_get_float_integer(args, &f)) < 0)
	return r;
      switch (type) {
      case FUNCTION_TYPE_PI:     f = M_PI * f;                 break;
      case FUNCTION_TYPE_RADIAN: f = (2.0 * M_PI) / 360.0 * f; break;
      case FUNCTION_TYPE_DEGREE: f = 360.0 / (2.0 * M_PI) * f; break;
      case FUNCTION_TYPE_SIN:    f = sin(f);                   break;
      case FUNCTION_TYPE_COS:    f = cos(f);                   break;
      case FUNCTION_TYPE_TAN:    f = tan(f);                   break;
      case FUNCTION_TYPE_ASIN:   f = asin(f);                  break;
      case FUNCTION_TYPE_ACOS:   f = acos(f);                  break;
      case FUNCTION_TYPE_ATAN:   f = atan(f);                  break;
      case FUNCTION_TYPE_SINH:   f = sinh(f);                  break;
      case FUNCTION_TYPE_COSH:   f = cosh(f);                  break;
      case FUNCTION_TYPE_TANH:   f = tanh(f);                  break;
      case FUNCTION_TYPE_EXP:    f = exp(f);                   break;
      case FUNCTION_TYPE_LOG:    f = log(f);                   break;
      case FUNCTION_TYPE_LOG10:  f = log10(f);                 break;
      case FUNCTION_TYPE_SQRT:   f = sqrt(f);                  break;
      default:
	return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
      }
      if (value) {
	if ((r = value_set_float(value, f)) < 0)
	  return r;
	value = value->next;
      }
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
#endif
}

static int function_pi(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_PI, value, args);
}

static int function_radian(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_RADIAN, value, args);
}

static int function_degree(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_DEGREE, value, args);
}

static int function_sin(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_SIN, value, args);
}

static int function_cos(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_COS, value, args);
}

static int function_tan(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_TAN, value, args);
}

static int function_asin(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_ASIN, value, args);
}

static int function_acos(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_ACOS, value, args);
}

static int function_atan(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_ATAN, value, args);
}

static int function_sinh(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_SINH, value, args);
}

static int function_cosh(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_COSH, value, args);
}

static int function_tanh(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_TANH, value, args);
}

static int function_exp(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_EXP, value, args);
}

static int function_log(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_LOG, value, args);
}

static int function_log10(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_LOG10, value, args);
}

static int function_sqrt(value_t value, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_SQRT, value, args);
}

static int function_mathfunc2(function_type_t type, value_t value, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  double f0, f1;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_float_integer(args, &f0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_float_integer(args, &f1)) < 0)
    return r;

  switch (type) {
  case FUNCTION_TYPE_ATAN2: f0 = atan2(f0, f1); break;
  case FUNCTION_TYPE_POW:   f0 = pow(f0, f1);   break;
  case FUNCTION_TYPE_FMOD:  f0 = fmod(f0, f1);  break;
  default:
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_float(value, f0)) < 0)
    return r;

  return 0;
#endif
}

static int function_atan2(value_t value, value_t args)
{
  return function_mathfunc2(FUNCTION_TYPE_ATAN2, value, args);
}

static int function_pow(value_t value, value_t args)
{
  return function_mathfunc2(FUNCTION_TYPE_POW, value, args);
}

static int function_fmod(value_t value, value_t args)
{
  return function_mathfunc2(FUNCTION_TYPE_FMOD, value, args);
}

static int function_mathvalue(function_type_t type, value_t value, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  double f;

  if (!args) {
    switch (type) {
#ifdef USE_MATHBUILTIN
    case FUNCTION_TYPE_INFINITY: f = INFINITY; break;
    case FUNCTION_TYPE_NAN:      f = NAN;      break;
#endif
    default:
      return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
    }
    if (value) {
      if ((r = value_set_float(value, f)) < 0)
	return r;
      value = value->next;
    }
  } else {
    for (; args; args = args->next) {
      switch (type) {
#ifdef USE_MATHBUILTIN
      case FUNCTION_TYPE_INFINITY: f = INFINITY; break;
      case FUNCTION_TYPE_NAN:      f = NAN;      break;
#endif
      default:
	return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
      }
      if (value) {
	if ((r = value_set_float(value, f)) < 0)
	  return r;
	value = value->next;
      }
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
#endif
}

static int function_infinity(value_t value, value_t args)
{
  return function_mathvalue(FUNCTION_TYPE_INFINITY, value, args);
}

static int function_nan(value_t value, value_t args)
{
  return function_mathvalue(FUNCTION_TYPE_NAN, value, args);
}

static int function_mathcheck(function_type_t type, value_t value, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, is = 0;
  double f;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_float_integer(args, &f)) < 0)
      return r;
    switch (type) {
#ifdef USE_MATHBUILTIN
    case FUNCTION_TYPE_ISFINITE: is = isfinite(f) ? 1 : 0; break;
    case FUNCTION_TYPE_ISINF   : is = isinf(f)    ? 1 : 0; break;
    case FUNCTION_TYPE_ISNAN   : is = isnan(f)    ? 1 : 0; break;
    case FUNCTION_TYPE_ISNORMAL: is = isnormal(f) ? 1 : 0; break;
#endif
    default:
      return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
    }
    if (value) {
      if ((r = value_set_integer(value, is)) < 0)
	return r;
      value = value->next;
    }
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
#endif
}

static int function_isfinite(value_t value, value_t args)
{
  return function_mathcheck(FUNCTION_TYPE_ISFINITE, value, args);
}

static int function_isinf(value_t value, value_t args)
{
  return function_mathcheck(FUNCTION_TYPE_ISINF, value, args);
}

static int function_isnan(value_t value, value_t args)
{
  return function_mathcheck(FUNCTION_TYPE_ISNAN, value, args);
}

static int function_isnormal(value_t value, value_t args)
{
  return function_mathcheck(FUNCTION_TYPE_ISNORMAL, value, args);
}
#endif

static int function_fopen(value_t value, value_t args)
{
  int r;
  char *filename, *mode = "r";
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &filename, NULL)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_string(args, &mode, NULL)) < 0)
      ;
  }

  fp = fopen(filename, mode);
  if (!fp)
    return NLL_ERRCODE_FILE_NOT_FOUND;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, (integer_t)fp)) < 0)
    return r;

  return 0;
}

static int function_fclose(value_t value, value_t args)
{
  int r;
  integer_t integer;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  fp = (FILE *)integer;
  if (fclose(fp))
    return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, 0)) < 0)
    return r;

  return 0;
}

static int function_fgets(value_t value, value_t args)
{
  int r;
  integer_t integer;
  char buffer[STRING_MAXLEN + 1];
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  fp = integer ? (FILE *)integer : nll_stdin;
  if (!fgets(buffer, sizeof(buffer), fp))
    return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, -1)) < 0)
    return r;

  return 0;
}

static int function_fputs(value_t value, value_t args)
{
  int r, length;
  integer_t integer;
  char *string;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &string, &length)) < 0)
    return r;

  fp = integer ? (FILE *)integer : nll_stdout;
  if (!fputs(string, fp))
    return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;
  fflush(fp);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, length)) < 0)
    return r;

  return 0;
}

static int function_fread(value_t value, value_t args)
{
  int r;
  area_t area;
  integer_t integer, offset = 0, size = -1;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
  }

  if (size < 0)
    size = area_get_size(area) - offset;

  if (size < 0)
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  if ((offset < 0) || (offset + size > area_get_size(area)))
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  fp = integer ? (FILE *)integer : nll_stdin;
  if (fread(area_get_buffer(area) + offset, 1, size, fp) != size)
    return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, size)) < 0)
    return r;

  return 0;
}

static int function_fwrite(value_t value, value_t args)
{
  int r;
  area_t area;
  integer_t integer, offset = 0, size = -1;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
  }

  if (size < 0)
    size = area_get_size(area) - offset;

  if (size < 0)
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  if ((offset < 0) || (offset + size > area_get_size(area)))
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  fp = integer ? (FILE *)integer : nll_stdout;
  if (fwrite(area_get_buffer(area) + offset, 1, size, fp) != size)
    return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;
  fflush(fp);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, size)) < 0)
    return r;

  return 0;
}

static int function_keyname(value_t value, value_t args)
{
  int r;
  integer_t keycode;
  char *keyname;

  for (; args && value; args = args->next, value = value->next) {
    if ((r = value_get_integer(args, &keycode)) < 0)
      continue;
    keyname = keyname_get(keycode);
    if (!keyname) keyname = "";
    if ((r = value_set_string(value, keyname, -1)) < 0)
      return r;
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_keycode(value_t value, value_t args)
{
  int r;
  char *keyname;
  key_code_t keycode;

  for (; args && value; args = args->next, value = value->next) {
    if ((r = value_get_string(args, &keyname, NULL)) < 0)
      continue;
    keycode = keycode_get(keyname);
    if ((r = value_set_integer(value, keycode)) < 0)
      return r;
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int function_screen(value_t value, value_t args)
{
  int r;
  integer_t mode = -1, width = -1, height = -1, flags = 0;

  if (args) {
    if ((r = value_get_integer(args, &mode)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &width)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &height)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
  }

  r = framebuf_setscreen(mode, width, height, flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
}

static int function_scrstat(value_t value, value_t args)
{
  int r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_screen_mode())) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_screen_width())) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_screen_height())) < 0)
    return r;

  return 0;
}

static int function_flush(value_t value, value_t args)
{
  int r;
  integer_t flags = 0;

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
  }

  r = framebuf_flush(flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
}

static int function_getcolor(value_t value, value_t args)
{
  int r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_cursor_fgcolor())) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_cursor_bgcolor())) < 0)
    return r;

  return 0;
}

static int function_setcolor(value_t value, value_t args)
{
  int r;
  integer_t fgcolor = -1, bgcolor = -1;

  if (args) {
    if ((r = value_get_integer(args, &fgcolor)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &bgcolor)) < 0)
      ;
  }

  framebuf_setfgcolor(fgcolor);
  framebuf_setbgcolor(bgcolor);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, 0)) < 0)
    return r;

  return 0;
}

static int function_getattr(value_t value, value_t args)
{
  int r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_cursor_effect())) < 0)
    return r;

  return 0;
}

static int function_setattr(value_t value, value_t args)
{
  int r;
  integer_t effect = -1;

  if (args) {
    if ((r = value_get_integer(args, &effect)) < 0)
      ;
  }

  r = framebuf_seteffect(effect);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
}

static int function_getpos(value_t value, value_t args)
{
  int r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_cursor_x())) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_cursor_y())) < 0)
    return r;

  return 0;
}

static int function_move(value_t value, value_t args)
{
  int r;
  integer_t x = -1, y = -1, dx = 0, dy = 0;

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &dx)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &dy)) < 0)
      ;
  }

  framebuf_move_cursor(x, y);

  if      (dx < 0) framebuf_move_left(-dx);
  else if (dx > 0) framebuf_move_right(dx);

  if      (dy < 0) framebuf_move_up(-dy);
  else if (dy > 0) framebuf_move_down(dy);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, 0)) < 0)
    return r;

  return 0;
}

static int function_getch(value_t value, value_t args)
{
  int r;
  integer_t x = -1, y = -1;
  int ch;
  char s[2];

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
  }

  ch = framebuf_getch(x, y);

  s[0] = ch;
  s[1] = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, s, -1)) < 0)
    return r;

  return 0;
}

static int function_clear(value_t value, value_t args)
{
  int r;
  integer_t ch = ' ', flags = 0;

  if (args) {
    if ((r = value_get_integer(args, &ch)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
  }

  r = framebuf_clear(ch, flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
}

static int function_scroll(value_t value, value_t args)
{
  int r;
  integer_t x = 0, y = 0, ch = -1;

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &ch)) < 0)
      ;
  }

  framebuf_scroll(x, y, ch);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, 0)) < 0)
    return r;

  return 0;
}

static int function_line(value_t value, value_t args)
{
  int r;
  integer_t x0, y0, x1, y1;
  char *string = "";

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x1)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y1)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_string(args, &string, NULL)) < 0)
      ;
  }

  r = framebuf_line(x0, y0, x1, y1, string);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
}

static int function_box(value_t value, value_t args)
{
  int r;
  integer_t x0, y0, x1, y1, flags = 0;
  char *string = "";

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x1)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y1)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_string(args, &string, NULL)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
  }

  r = framebuf_box(x0, y0, x1, y1, string, flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
}

static int function_inkey(value_t value, value_t args)
{
  int r;
  integer_t integer;
  char key[2], *p;

  integer = framebuf_getkey();

  switch (integer) {
  case KEY_CODE_END:
  case KEY_CODE_UP:
  case KEY_CODE_DOWN:
  case KEY_CODE_RIGHT:
  case KEY_CODE_LEFT:
  case KEY_CODE_PAGEUP:
  case KEY_CODE_PAGEDOWN:
    p = keyname_get(integer);
    break;
  default:
    key[0] = integer;
    key[1] = '\0';
    p = key;
    break;
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, p, -1)) < 0)
    return r;

  return 0;
}

static int function_gscreen(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t mode = -1, width = -1, height = -1, flags = -1, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &mode)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &width)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &height)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_setscreen(mode, width, height, index, flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gscrstat(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, graphic_get_screen_mode())) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, graphic_get_screen_width())) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, graphic_get_screen_height())) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, graphic_get_screen_image())) < 0)
    return r;

  return 0;
#endif
}

static int function_gcolor(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t red = 0, green = 0, blue = 0;
  int color;

  if (args) {
    if ((r = value_get_integer(args, &red)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &green)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &blue)) < 0)
      ;
  }

  color = graphic_color(red, green, blue);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, color)) < 0)
    return r;

  return 0;
#endif
}

static int function_grgb(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t color = 0;
  int red, green, blue;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
  }

  graphic_rgb(color, &red, &green, &blue);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, red)) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, green)) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, blue)) < 0)
    return r;

  return 0;
#endif
}

static int function_gsync(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;

  r = graphic_sync();

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gflush(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t flags = 0;

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
  }

  r = graphic_flush(flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_ggetsize(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t index = NLL_G_MAINIMG;
  int width, height;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  width  = graphic_getwidth(index);
  height = graphic_getheight(index);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, width)) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, height)) < 0)
    return r;

  return 0;
#endif
}

static int function_ggetcolor(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_getcolor(index);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gsetcolor(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t color = 0, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_setcolor(index, color);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_ggetpixel(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t x = -1, y = -1, index = NLL_G_MAINIMG;
  int color;

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  color = graphic_getpixel(index, x, y);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, color)) < 0)
    return r;

  return 0;
#endif
}

static int function_gclear(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t color = 0, flags = 0, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_clear(index, color, flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gscroll(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t x = 0, y = 0, color = 0, flags = 0, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_scroll(index, x, y, color, flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gdot(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t x, y, color = -1, flags = 0, index = NLL_G_MAINIMG;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_dot(index, x, y, color, flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gline(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t x0, y0, x1, y1, color = -1, flags = 0, index = NLL_G_MAINIMG;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x1)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y1)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_line(index, x0, y0, x1, y1, color, flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gbox(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t x0, y0, x1, y1, color = -1, flags = 0, index = NLL_G_MAINIMG;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x1)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y1)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_box(index, x0, y0, x1, y1, color, flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gcircle(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t x, y, radius, color = -1, flags = 0, index = NLL_G_MAINIMG;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &radius)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_circle(index, x, y, radius, color, flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gcopyimage(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t x0 = 0, y0 = 0, source = NLL_G_MAINIMG, x = 0, y = 0, width = 0, height = 0, transparent = -1, flags = 0, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer_float(args, &x0)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y0)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &source)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &width)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &height)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &transparent)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_image(index, x0, y0, source, x, y, width, height, transparent, flags);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_ginkey(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, keycode;
  char *keyname;

  keycode = graphic_getkeybuf();
  keyname = keyname_get(keycode);
  if (!keyname) keyname = "";

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, keyname, -1)) < 0)
    return r;

  return 0;
#endif
}

static int function_gkeystat(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  char *keyname;
  int keycode, s;

  for (; args && value; args = args->next, value = value->next) {
    if ((r = value_get_string(args, &keyname, NULL)) < 0)
      continue;
    keycode = keycode_get(keyname);
    s = graphic_getkeystatus(keycode);
    if ((r = value_set_integer(value, s)) < 0)
      return r;
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
#endif
}

static int function_gmousepos(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, graphic_getmousex())) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, graphic_getmousey())) < 0)
    return r;

  return 0;
#endif
}

static int function_ginmouse(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, mousecode, x, y;
  char *mousename;

  mousecode = graphic_getmousebuf(&x, &y);
  mousename = mousename_get(mousecode);
  if (!mousename) {
    mousename = "";
    x = -1;
    y = -1;
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, mousename, -1)) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, x)) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, y)) < 0)
    return r;

  return 0;
#endif
}

static int function_gmousestat(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  char *mousename;
  int mousecode, s;

  for (; args && value; args = args->next, value = value->next) {
    if ((r = value_get_string(args, &mousename, NULL)) < 0)
      continue;
    mousecode = mousecode_get(mousename);
    s = graphic_getmousestatus(mousecode);
    if ((r = value_set_integer(value, s)) < 0)
      return r;
  }

  for (; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
#endif
}

static int function_gmainimage(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_mainimage(index);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gallocimage(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t width = 0, height = 0;

  if (args) {
    if ((r = value_get_integer_float(args, &width)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &height)) < 0)
      ;
  }

  r = graphic_allocimage(width, height);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gfreeimage(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  r = graphic_freeimage(index);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}

static int function_gloadimage(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, width = 0, height = 0;
  char *filename;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &filename, NULL)) < 0)
    return r;

  r = graphic_loadimage(filename, &width, &height);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, width)) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, height)) < 0)
    return r;

  return 0;
#endif
}

static int function_gsaveimage(value_t value, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t source = NLL_G_MAINIMG, x = 0, y = 0, width = 0, height = 0;
  char *filename;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &filename, NULL)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &source)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &width)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &height)) < 0)
      ;
  }

  r = graphic_saveimage(filename, source, x, y, width, height);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, r)) < 0)
    return r;

  return 0;
#endif
}
