#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#ifdef USE_MATHLIB
#include <math.h>
#endif
#ifndef WIN32
#include <sys/socket.h>
#else
#include <winsock2.h>
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "const.h"
#include "nlllib.h"
#include "memory.h"
#include "value.h"
#include "string.h"
#include "variable.h"
#include "key.h"

static struct const_variable {
  char *name;
  value_type_t type;
  integer_t integer;
  char *string;
  void *pointer;
} const_variables[] = {
  { "NULL"        , VALUE_TYPE_POINTER  , 0               , NULL, NULL },

  { "D_VARIABLE"  , VALUE_TYPE_INTEGER  , NLL_D_VARIABLE  , NULL, NULL },
  { "D_STRING"    , VALUE_TYPE_INTEGER  , NLL_D_STRING    , NULL, NULL },
  { "D_LABEL"     , VALUE_TYPE_INTEGER  , NLL_D_LABEL     , NULL, NULL },
  { "D_LINE"      , VALUE_TYPE_INTEGER  , NLL_D_LINE      , NULL, NULL },
  { "D_STACK"     , VALUE_TYPE_INTEGER  , NLL_D_STACK     , NULL, NULL },
  { "D_ALL"       , VALUE_TYPE_INTEGER  , NLL_D_ALL       , NULL, NULL },
  { "L_COMMAND"   , VALUE_TYPE_INTEGER  , NLL_L_COMMAND   , NULL, NULL },
  { "L_FUNCTION"  , VALUE_TYPE_INTEGER  , NLL_L_FUNCTION  , NULL, NULL },
  { "L_CONST"     , VALUE_TYPE_INTEGER  , NLL_L_CONST     , NULL, NULL },

  { "F_DISABLE"   , VALUE_TYPE_INTEGER  , NLL_F_DISABLE   , NULL, NULL },
  { "F_BUFFER"    , VALUE_TYPE_INTEGER  , NLL_F_BUFFER    , NULL, NULL },
  { "F_AUTOMODE"  , VALUE_TYPE_INTEGER  , NLL_F_AUTOMODE  , NULL, NULL },
  { "F_FLUSH"     , VALUE_TYPE_INTEGER  , NLL_F_FLUSH     , NULL, NULL },
  { "F_CURSOR"    , VALUE_TYPE_INTEGER  , NLL_F_CURSOR    , NULL, NULL },
  { "F_CURSES"    , VALUE_TYPE_INTEGER  , NLL_F_CURSES    , NULL, NULL },

  { "F_FILL"      , VALUE_TYPE_INTEGER  , NLL_F_FILL      , NULL, NULL },

  { "C_RESET"     , VALUE_TYPE_INTEGER  , NLL_C_RESET     , NULL, NULL },
  { "C_BLACK"     , VALUE_TYPE_INTEGER  , NLL_C_BLACK     , NULL, NULL },
  { "C_RED"       , VALUE_TYPE_INTEGER  , NLL_C_RED       , NULL, NULL },
  { "C_GREEN"     , VALUE_TYPE_INTEGER  , NLL_C_GREEN     , NULL, NULL },
  { "C_YELLOW"    , VALUE_TYPE_INTEGER  , NLL_C_YELLOW    , NULL, NULL },
  { "C_BLUE"      , VALUE_TYPE_INTEGER  , NLL_C_BLUE      , NULL, NULL },
  { "C_MAGENTA"   , VALUE_TYPE_INTEGER  , NLL_C_MAGENTA   , NULL, NULL },
  { "C_CYAN"      , VALUE_TYPE_INTEGER  , NLL_C_CYAN      , NULL, NULL },
  { "C_WHITE"     , VALUE_TYPE_INTEGER  , NLL_C_WHITE     , NULL, NULL },

  { "A_RESET"     , VALUE_TYPE_INTEGER  , NLL_A_RESET     , NULL, NULL },
  { "A_BOLD"      , VALUE_TYPE_INTEGER  , NLL_A_BOLD      , NULL, NULL },
  { "A_FAINT"     , VALUE_TYPE_INTEGER  , NLL_A_FAINT     , NULL, NULL },
  { "A_ITALIC"    , VALUE_TYPE_INTEGER  , NLL_A_ITALIC    , NULL, NULL },
  { "A_UNDERLINE" , VALUE_TYPE_INTEGER  , NLL_A_UNDERLINE , NULL, NULL },
  { "A_BLINK"     , VALUE_TYPE_INTEGER  , NLL_A_BLINK     , NULL, NULL },
  { "A_RAPIDBLINK", VALUE_TYPE_INTEGER  , NLL_A_RAPIDBLINK, NULL, NULL },
  { "A_INVERSE"   , VALUE_TYPE_INTEGER  , NLL_A_INVERSE   , NULL, NULL },
  { "A_INVISIBLE" , VALUE_TYPE_INTEGER  , NLL_A_INVISIBLE , NULL, NULL },
  { "A_CROSSOUT"  , VALUE_TYPE_INTEGER  , NLL_A_CROSSOUT  , NULL, NULL },

  { "KEY_ENTER"       , VALUE_TYPE_INTEGER  , KEY_CODE_ENTER       , NULL, NULL },
  { "KEY_ESCAPE"      , VALUE_TYPE_INTEGER  , KEY_CODE_ESCAPE      , NULL, NULL },
  { "KEY_BACKSPACE"   , VALUE_TYPE_INTEGER  , KEY_CODE_BACKSPACE   , NULL, NULL },
  { "KEY_TAB"         , VALUE_TYPE_INTEGER  , KEY_CODE_TAB         , NULL, NULL },
  { "KEY_SPACE"       , VALUE_TYPE_INTEGER  , KEY_CODE_SPACE       , NULL, NULL },
  { "KEY_EXCLAIM"     , VALUE_TYPE_INTEGER  , KEY_CODE_EXCLAIM     , NULL, NULL },
  { "KEY_QUOTEDBL"    , VALUE_TYPE_INTEGER  , KEY_CODE_QUOTEDBL    , NULL, NULL },
  { "KEY_HASH"        , VALUE_TYPE_INTEGER  , KEY_CODE_HASH        , NULL, NULL },
  { "KEY_PERCENT"     , VALUE_TYPE_INTEGER  , KEY_CODE_PERCENT     , NULL, NULL },
  { "KEY_DOLLAR"      , VALUE_TYPE_INTEGER  , KEY_CODE_DOLLAR      , NULL, NULL },
  { "KEY_AMPERSAND"   , VALUE_TYPE_INTEGER  , KEY_CODE_AMPERSAND   , NULL, NULL },
  { "KEY_QUOTE"       , VALUE_TYPE_INTEGER  , KEY_CODE_QUOTE       , NULL, NULL },
  { "KEY_LEFTPAREN"   , VALUE_TYPE_INTEGER  , KEY_CODE_LEFTPAREN   , NULL, NULL },
  { "KEY_RIGHTPAREN"  , VALUE_TYPE_INTEGER  , KEY_CODE_RIGHTPAREN  , NULL, NULL },
  { "KEY_ASTERISK"    , VALUE_TYPE_INTEGER  , KEY_CODE_ASTERISK    , NULL, NULL },
  { "KEY_PLUS"        , VALUE_TYPE_INTEGER  , KEY_CODE_PLUS        , NULL, NULL },
  { "KEY_COMMA"       , VALUE_TYPE_INTEGER  , KEY_CODE_COMMA       , NULL, NULL },
  { "KEY_MINUS"       , VALUE_TYPE_INTEGER  , KEY_CODE_MINUS       , NULL, NULL },
  { "KEY_PERIOD"      , VALUE_TYPE_INTEGER  , KEY_CODE_PERIOD      , NULL, NULL },
  { "KEY_SLASH"       , VALUE_TYPE_INTEGER  , KEY_CODE_SLASH       , NULL, NULL },
  { "KEY_0"           , VALUE_TYPE_INTEGER  , KEY_CODE_0           , NULL, NULL },
  { "KEY_1"           , VALUE_TYPE_INTEGER  , KEY_CODE_1           , NULL, NULL },
  { "KEY_2"           , VALUE_TYPE_INTEGER  , KEY_CODE_2           , NULL, NULL },
  { "KEY_3"           , VALUE_TYPE_INTEGER  , KEY_CODE_3           , NULL, NULL },
  { "KEY_4"           , VALUE_TYPE_INTEGER  , KEY_CODE_4           , NULL, NULL },
  { "KEY_5"           , VALUE_TYPE_INTEGER  , KEY_CODE_5           , NULL, NULL },
  { "KEY_6"           , VALUE_TYPE_INTEGER  , KEY_CODE_6           , NULL, NULL },
  { "KEY_7"           , VALUE_TYPE_INTEGER  , KEY_CODE_7           , NULL, NULL },
  { "KEY_8"           , VALUE_TYPE_INTEGER  , KEY_CODE_8           , NULL, NULL },
  { "KEY_9"           , VALUE_TYPE_INTEGER  , KEY_CODE_9           , NULL, NULL },
  { "KEY_COLON"       , VALUE_TYPE_INTEGER  , KEY_CODE_COLON       , NULL, NULL },
  { "KEY_SEMICOLON"   , VALUE_TYPE_INTEGER  , KEY_CODE_SEMICOLON   , NULL, NULL },
  { "KEY_LESS"        , VALUE_TYPE_INTEGER  , KEY_CODE_LESS        , NULL, NULL },
  { "KEY_EQUALS"      , VALUE_TYPE_INTEGER  , KEY_CODE_EQUALS      , NULL, NULL },
  { "KEY_GREATER"     , VALUE_TYPE_INTEGER  , KEY_CODE_GREATER     , NULL, NULL },
  { "KEY_QUESTION"    , VALUE_TYPE_INTEGER  , KEY_CODE_QUESTION    , NULL, NULL },
  { "KEY_AT"          , VALUE_TYPE_INTEGER  , KEY_CODE_AT          , NULL, NULL },
  { "KEY_LEFTBRACKET" , VALUE_TYPE_INTEGER  , KEY_CODE_LEFTBRACKET , NULL, NULL },
  { "KEY_BACKSLASH"   , VALUE_TYPE_INTEGER  , KEY_CODE_BACKSLASH   , NULL, NULL },
  { "KEY_RIGHTBRACKET", VALUE_TYPE_INTEGER  , KEY_CODE_RIGHTBRACKET, NULL, NULL },
  { "KEY_CARET"       , VALUE_TYPE_INTEGER  , KEY_CODE_CARET       , NULL, NULL },
  { "KEY_UNDERSCORE"  , VALUE_TYPE_INTEGER  , KEY_CODE_UNDERSCORE  , NULL, NULL },
  { "KEY_BACKQUOTE"   , VALUE_TYPE_INTEGER  , KEY_CODE_BACKQUOTE   , NULL, NULL },
  { "KEY_A"           , VALUE_TYPE_INTEGER  , KEY_CODE_A           , NULL, NULL },
  { "KEY_B"           , VALUE_TYPE_INTEGER  , KEY_CODE_B           , NULL, NULL },
  { "KEY_C"           , VALUE_TYPE_INTEGER  , KEY_CODE_C           , NULL, NULL },
  { "KEY_D"           , VALUE_TYPE_INTEGER  , KEY_CODE_D           , NULL, NULL },
  { "KEY_E"           , VALUE_TYPE_INTEGER  , KEY_CODE_E           , NULL, NULL },
  { "KEY_F"           , VALUE_TYPE_INTEGER  , KEY_CODE_F           , NULL, NULL },
  { "KEY_G"           , VALUE_TYPE_INTEGER  , KEY_CODE_G           , NULL, NULL },
  { "KEY_H"           , VALUE_TYPE_INTEGER  , KEY_CODE_H           , NULL, NULL },
  { "KEY_I"           , VALUE_TYPE_INTEGER  , KEY_CODE_I           , NULL, NULL },
  { "KEY_J"           , VALUE_TYPE_INTEGER  , KEY_CODE_J           , NULL, NULL },
  { "KEY_K"           , VALUE_TYPE_INTEGER  , KEY_CODE_K           , NULL, NULL },
  { "KEY_L"           , VALUE_TYPE_INTEGER  , KEY_CODE_L           , NULL, NULL },
  { "KEY_M"           , VALUE_TYPE_INTEGER  , KEY_CODE_M           , NULL, NULL },
  { "KEY_N"           , VALUE_TYPE_INTEGER  , KEY_CODE_N           , NULL, NULL },
  { "KEY_O"           , VALUE_TYPE_INTEGER  , KEY_CODE_O           , NULL, NULL },
  { "KEY_P"           , VALUE_TYPE_INTEGER  , KEY_CODE_P           , NULL, NULL },
  { "KEY_Q"           , VALUE_TYPE_INTEGER  , KEY_CODE_Q           , NULL, NULL },
  { "KEY_R"           , VALUE_TYPE_INTEGER  , KEY_CODE_R           , NULL, NULL },
  { "KEY_S"           , VALUE_TYPE_INTEGER  , KEY_CODE_S           , NULL, NULL },
  { "KEY_T"           , VALUE_TYPE_INTEGER  , KEY_CODE_T           , NULL, NULL },
  { "KEY_U"           , VALUE_TYPE_INTEGER  , KEY_CODE_U           , NULL, NULL },
  { "KEY_V"           , VALUE_TYPE_INTEGER  , KEY_CODE_V           , NULL, NULL },
  { "KEY_W"           , VALUE_TYPE_INTEGER  , KEY_CODE_W           , NULL, NULL },
  { "KEY_X"           , VALUE_TYPE_INTEGER  , KEY_CODE_X           , NULL, NULL },
  { "KEY_Y"           , VALUE_TYPE_INTEGER  , KEY_CODE_Y           , NULL, NULL },
  { "KEY_Z"           , VALUE_TYPE_INTEGER  , KEY_CODE_Z           , NULL, NULL },

  { "KEY_CAPSLOCK"    , VALUE_TYPE_INTEGER  , KEY_CODE_CAPSLOCK    , NULL, NULL },
  { "KEY_F1"          , VALUE_TYPE_INTEGER  , KEY_CODE_F1          , NULL, NULL },
  { "KEY_F2"          , VALUE_TYPE_INTEGER  , KEY_CODE_F2          , NULL, NULL },
  { "KEY_F3"          , VALUE_TYPE_INTEGER  , KEY_CODE_F3          , NULL, NULL },
  { "KEY_F4"          , VALUE_TYPE_INTEGER  , KEY_CODE_F4          , NULL, NULL },
  { "KEY_F5"          , VALUE_TYPE_INTEGER  , KEY_CODE_F5          , NULL, NULL },
  { "KEY_F6"          , VALUE_TYPE_INTEGER  , KEY_CODE_F6          , NULL, NULL },
  { "KEY_F7"          , VALUE_TYPE_INTEGER  , KEY_CODE_F7          , NULL, NULL },
  { "KEY_F8"          , VALUE_TYPE_INTEGER  , KEY_CODE_F8          , NULL, NULL },
  { "KEY_F9"          , VALUE_TYPE_INTEGER  , KEY_CODE_F9          , NULL, NULL },
  { "KEY_F10"         , VALUE_TYPE_INTEGER  , KEY_CODE_F10         , NULL, NULL },
  { "KEY_F11"         , VALUE_TYPE_INTEGER  , KEY_CODE_F11         , NULL, NULL },
  { "KEY_F12"         , VALUE_TYPE_INTEGER  , KEY_CODE_F12         , NULL, NULL },

  { "KEY_PRINTSCREEN" , VALUE_TYPE_INTEGER  , KEY_CODE_PRINTSCREEN , NULL, NULL },
  { "KEY_SCROLLLOCK"  , VALUE_TYPE_INTEGER  , KEY_CODE_SCROLLLOCK  , NULL, NULL },
  { "KEY_PAUSE"       , VALUE_TYPE_INTEGER  , KEY_CODE_PAUSE       , NULL, NULL },
  { "KEY_INSERT"      , VALUE_TYPE_INTEGER  , KEY_CODE_INSERT      , NULL, NULL },
  { "KEY_HOME"        , VALUE_TYPE_INTEGER  , KEY_CODE_HOME        , NULL, NULL },
  { "KEY_PAGEUP"      , VALUE_TYPE_INTEGER  , KEY_CODE_PAGEUP      , NULL, NULL },
  { "KEY_DELETE"      , VALUE_TYPE_INTEGER  , KEY_CODE_DELETE      , NULL, NULL },
  { "KEY_END"         , VALUE_TYPE_INTEGER  , KEY_CODE_END         , NULL, NULL },
  { "KEY_PAGEDOWN"    , VALUE_TYPE_INTEGER  , KEY_CODE_PAGEDOWN    , NULL, NULL },
  { "KEY_RIGHT"       , VALUE_TYPE_INTEGER  , KEY_CODE_RIGHT       , NULL, NULL },
  { "KEY_LEFT"        , VALUE_TYPE_INTEGER  , KEY_CODE_LEFT        , NULL, NULL },
  { "KEY_DOWN"        , VALUE_TYPE_INTEGER  , KEY_CODE_DOWN        , NULL, NULL },
  { "KEY_UP"          , VALUE_TYPE_INTEGER  , KEY_CODE_UP          , NULL, NULL },

  { "KEY_F13"         , VALUE_TYPE_INTEGER  , KEY_CODE_F13         , NULL, NULL },
  { "KEY_F14"         , VALUE_TYPE_INTEGER  , KEY_CODE_F14         , NULL, NULL },
  { "KEY_F15"         , VALUE_TYPE_INTEGER  , KEY_CODE_F15         , NULL, NULL },

  { "KEY_LCTRL"       , VALUE_TYPE_INTEGER  , KEY_CODE_LCTRL       , NULL, NULL },
  { "KEY_LSHIFT"      , VALUE_TYPE_INTEGER  , KEY_CODE_LSHIFT      , NULL, NULL },
  { "KEY_LALT"        , VALUE_TYPE_INTEGER  , KEY_CODE_LALT        , NULL, NULL },
  { "KEY_RCTRL"       , VALUE_TYPE_INTEGER  , KEY_CODE_RCTRL       , NULL, NULL },
  { "KEY_RSHIFT"      , VALUE_TYPE_INTEGER  , KEY_CODE_RSHIFT      , NULL, NULL },
  { "KEY_RALT"        , VALUE_TYPE_INTEGER  , KEY_CODE_RALT        , NULL, NULL },

#ifdef O_RDONLY
  { "O_RDONLY"    , VALUE_TYPE_INTEGER  , O_RDONLY        , NULL, NULL },
#endif
#ifdef O_WRONLY
  { "O_WRONLY"    , VALUE_TYPE_INTEGER  , O_WRONLY        , NULL, NULL },
#endif
#ifdef O_RDWR
  { "O_RDWR"      , VALUE_TYPE_INTEGER  , O_RDWR          , NULL, NULL },
#endif

#ifdef AF_INET
  { "AF_INET"     , VALUE_TYPE_INTEGER  , AF_INET         , NULL, NULL },
#endif
#ifdef SOCK_STREAM
  { "SOCK_STREAM" , VALUE_TYPE_INTEGER  , SOCK_STREAM     , NULL, NULL },
#endif
#ifdef SOCK_DGRAM
  { "SOCK_DGRAM"  , VALUE_TYPE_INTEGER  , SOCK_DGRAM      , NULL, NULL },
#endif

#ifdef NLL_FLOATING_POINT
#ifdef USE_MATHLIB
  { "M_E"         , VALUE_TYPE_FLOAT    , 0               , NULL, NULL },
  { "M_PI"        , VALUE_TYPE_FLOAT    , 0               , NULL, NULL },
#endif
#endif

#ifdef USE_GRAPHIC
  { "G_DISABLE"   , VALUE_TYPE_INTEGER  , NLL_G_DISABLE   , NULL, NULL },
  { "G_DIRECT"    , VALUE_TYPE_INTEGER  , NLL_G_DIRECT    , NULL, NULL },
  { "G_BUFFER"    , VALUE_TYPE_INTEGER  , NLL_G_BUFFER    , NULL, NULL },
  { "G_AUTOMODE"  , VALUE_TYPE_INTEGER  , NLL_G_AUTOMODE  , NULL, NULL },
  { "G_FLUSH"     , VALUE_TYPE_INTEGER  , NLL_G_FLUSH     , NULL, NULL },
  { "G_RENDERER"  , VALUE_TYPE_INTEGER  , NLL_G_RENDERER  , NULL, NULL },
  { "G_FORCE"     , VALUE_TYPE_INTEGER  , NLL_G_FORCE     , NULL, NULL },
  { "G_SYNC"      , VALUE_TYPE_INTEGER  , NLL_G_SYNC      , NULL, NULL },
  { "G_ROUGH"     , VALUE_TYPE_INTEGER  , NLL_G_ROUGH     , NULL, NULL },
  { "G_HROUGH"    , VALUE_TYPE_INTEGER  , NLL_G_HROUGH    , NULL, NULL },
  { "G_MAINIMG"   , VALUE_TYPE_INTEGER  , NLL_G_MAINIMG   , NULL, NULL },

  { "G_SET"       , VALUE_TYPE_INTEGER  , NLL_G_SET       , NULL, NULL },
  { "G_AND"       , VALUE_TYPE_INTEGER  , NLL_G_AND       , NULL, NULL },
  { "G_OR"        , VALUE_TYPE_INTEGER  , NLL_G_OR        , NULL, NULL },
  { "G_XOR"       , VALUE_TYPE_INTEGER  , NLL_G_XOR       , NULL, NULL },
  { "G_FILL"      , VALUE_TYPE_INTEGER  , NLL_G_FILL      , NULL, NULL },

  { "G_BLACK"     , VALUE_TYPE_INTEGER  , NLL_G_BLACK     , NULL, NULL },
  { "G_BLUE"      , VALUE_TYPE_INTEGER  , NLL_G_BLUE      , NULL, NULL },
  { "G_GREEN"     , VALUE_TYPE_INTEGER  , NLL_G_GREEN     , NULL, NULL },
  { "G_CYAN"      , VALUE_TYPE_INTEGER  , NLL_G_CYAN      , NULL, NULL },
  { "G_RED"       , VALUE_TYPE_INTEGER  , NLL_G_RED       , NULL, NULL },
  { "G_MAGENTA"   , VALUE_TYPE_INTEGER  , NLL_G_MAGENTA   , NULL, NULL },
  { "G_YELLOW"    , VALUE_TYPE_INTEGER  , NLL_G_YELLOW    , NULL, NULL },
  { "G_WHITE"     , VALUE_TYPE_INTEGER  , NLL_G_WHITE     , NULL, NULL },
#endif

  { NULL          , VALUE_TYPE_UNDEFINED, 0               , NULL, NULL }
};

#ifdef NLL_FLOATING_POINT
double const_variables_floating[] = {
#ifdef USE_MATHLIB
  M_E,
  M_PI,
#endif
  0.0
};
#endif

static int variable_num = 0;

static variable_t head = NULL;

#ifndef NLL_MEMORY_STATIC
static int _init(void)
{
  return 0;
}

static int _check(void)
{
  return 0;
}

static variable_t _alloc(void)
{
  variable_t variable;
  if ((variable = memory_alloc(sizeof(*variable))) == NULL)
    return NULL;
  return variable;
}

static int _free(variable_t variable)
{
  memory_free(variable);
  return 0;
}
#else
static struct variable variables[VARIABLE_MAXNUM];

static variable_t pool = NULL;

static int _init(void)
{
  variable_t variable;
  int i;

  pool = NULL;
  memset(variables, 0, sizeof(variables));

  for (i = 0; i < VARIABLE_MAXNUM; i++) {
    variable = &variables[i];
    variable->next = pool;
    pool = variable;
  }

  return 0;
}

static int _check(void)
{
  variable_t variable;
  int n = 0;

  for (variable = pool; variable && (n < VARIABLE_MAXNUM); variable = variable->next)
    n++;

  if (variable || (n != VARIABLE_MAXNUM))
    return NLL_ERRCODE_VARIABLE_NOT_EMPTY;

  return 0;
}

static variable_t _alloc(void)
{
  variable_t variable;

  if (!pool)
    return NULL;

  variable = pool;
  pool = pool->next;

  return variable;
}

static int _free(variable_t variable)
{
  variable->next = pool;
  pool = variable;
  return 0;
}
#endif

int variable_check(void)
{
  if (variable_num || head)
    return NLL_ERRCODE_VARIABLE_NOT_EMPTY;

  return _check();
}

static int variable_free(variable_t variable)
{
  variable_t next;
  int r;

  for (; variable; variable = next) {
    next = variable->next;

    if (variable->value)
      value_free(variable->value);

    variable_num--;
    if ((r = _free(variable)) < 0)
      return r;
  }

  return 0;
}

static int variable_alloc(variable_t *variablep, int num)
{
  variable_t variable;
  int r;

  if ((variable = _alloc()) == NULL)
    return NLL_ERRCODE_VARIABLE_BUFFER_OVER;
  variable_num++;

  memset(variable, 0, sizeof(*variable));

  if ((r = variable_extend_value(variable, num)) < 0) {
    variable_free(variable);
    return r;
  }

  *variablep = variable;

  return 0;
}

int variable_del(variable_t variable)
{
  variable_t *p, next;

  for (p = &head; *p;) {
    if (*p == variable) {
      if ((--(variable->refcount) == 0) &&
	  (variable->value->type == VALUE_TYPE_UNDEFINED)) {
	next = (*p)->next;
	(*p)->next = NULL;
	variable_free(*p);
	*p = next;
	continue;
      }
    }
    p = &((*p)->next);
  }

  return 0;
}

int variable_clear(void)
{
  variable_t *p, variable, next;

  for (p = &head; *p;) {
    variable = *p;
    if (variable->refcount == 0) {
      next = variable->next;
      variable->next = NULL;
      variable_free(variable);
      *p = next;
      continue;
    }
    p = &(variable->next);
  }

  return 0;
}

int variable_clean(variable_t variable)
{
  int r;
  if ((r = value_clear(variable->value)) < 0)
    return r;
  return 0;
}

int variable_init(void)
{
  variable_num = 0;
  head = NULL;
  return _init();
}

int variable_get_name(char *dname, const char *name)
{
  const char *p;

  for (p = name; *p; p++) {
    if (p == name) {
      if (isalpha(*p) || (*p == '_'))
	;
#ifdef NLL_MULTIBYTE_NAME
      else if (*p & 0x80)
	;
#endif
      else
	break;
    } else {
      if (isalnum(*p) || (*p == '_'))
	;
#ifdef NLL_MULTIBYTE_NAME
      else if (*p & 0x80)
	;
#endif
      else
	break;
    }
    if (p - name > VARIABLE_NAMELEN)
      return NLL_ERRCODE_VARIABLE_TOO_LONG;
    *(dname++) = toupper(*p);
  }

  *dname = '\0';

  return p - name;
}

static int variable_const(variable_t variable)
{
  int r;
  struct const_variable *p;
#ifdef NLL_FLOATING_POINT
  double *dp = const_variables_floating;
#endif

  for (p = const_variables; p->name; p++) {
    if (!strcmp(variable->name, p->name)) {
      switch (p->type) {
      case VALUE_TYPE_INTEGER:
	if ((r = value_set_integer(variable->value, p->integer)) < 0)
	  return r;
	break;
      case VALUE_TYPE_STRING:
	if ((r = value_set_string(variable->value, p->string, -1)) < 0)
	  return r;
	break;
#ifdef NLL_FLOATING_POINT
      case VALUE_TYPE_FLOAT:
	if ((r = value_set_float(variable->value, *dp)) < 0)
	  return r;
	break;
#endif
      case VALUE_TYPE_POINTER:
	if ((r = value_set_pointer(variable->value, p->pointer, 0)) < 0)
	  return r;
	break;
      default:
	return NLL_ERRCODE_VALUE_INVALID_TYPE;
      }
      variable->value->flags |= VALUE_FLAG_CONST;
    }
#ifdef NLL_FLOATING_POINT
    if (p->type == VALUE_TYPE_FLOAT)
      dp++;
#endif
  }

  return 0;
}

int variable_get(const char *name, variable_t *variablep)
{
  variable_t variable;
  int r;

  for (variable = head; variable; variable = variable->next) {
    if (variable->name[0] && !strcmp(variable->name, name))
      break;
  }

  if (!variable) {
    if ((r = variable_alloc(&variable, 1)) < 0)
      return r;

    strcpy(variable->name, name);

    if ((r = variable_const(variable)) < 0) {
      variable_free(variable);
      return r;
    }

    variable->next = head;
    head = variable;
  }

  variable->refcount++;

  *variablep = variable;

  return 0;
}

int variable_extend_value(variable_t variable, int num)
{
  value_t *valuep, value;
  int r, n = 0;

  for (valuep = &variable->value; *valuep; valuep = &(*valuep)->next)
    n++;

  while (n < num) {
    if ((r = value_alloc(&value)) < 0)
      return r;
    value->next = *valuep;
    *valuep = value;
    n++;
  }

  return 0;
}

int variable_dump(FILE *fp, variable_t variable)
{
  if (!variable)
    variable = head;

  nll_wait_output(fp);
  fprintf(fp, "\n-- Variable dump --\n");

  for (; variable; variable = variable->next) {
    nll_wait_output(fp);
    fprintf(fp, "%s:\tRef:%d\tValue:", variable->name, variable->refcount);
    value_dump(fp, variable->value);
    fprintf(fp, "\n");
  }

  fflush(fp);

  return 0;
}

int variable_const_list(FILE *fp)
{
  struct const_variable *p;
#ifdef NLL_FLOATING_POINT
  double *dp = const_variables_floating;
#endif

  for (p = const_variables; p->name; p++) {
    fprintf(fp, "%s", p->name);
    switch (p->type) {
    case VALUE_TYPE_INTEGER:
      fprintf(fp, "\t%ld(0x%lx)", p->integer, p->integer);
      break;
    case VALUE_TYPE_STRING:
      fprintf(fp, "\t\"%s\"", p->string);
      break;
#ifdef NLL_FLOATING_POINT
    case VALUE_TYPE_FLOAT:
      fprintf(fp, "\t%f(FLOAT)", *dp);
      break;
#endif
    case VALUE_TYPE_POINTER:
      fprintf(fp, "\t(POINTER)");
      break;
    default:
      return NLL_ERRCODE_VALUE_INVALID_TYPE;
    }
    fprintf(fp, "\n");
#ifdef NLL_FLOATING_POINT
    if (p->type == VALUE_TYPE_FLOAT)
      dp++;
#endif
  }

  return 0;
}
